"""Implementation of a style transfer network

References:
    - Gatys et al. - A Neural Algorithm of Artistic Style - https://arxiv.org/abs/1508.06576
    - Pytorch Neural Style tutorial: https://pytorch.org/tutorials/advanced/neural_style_tutorial.html
    - Torch implementation by jcjohnson https://github.com/jcjohnson/neural-style
    - Pytorch implementation by ProGamerGov https://github.com/ProGamerGov/neural-style-pt
"""
from collections.abc import Iterable
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models



def get_device():
    """Returns the Pytorch device into which to run the calculations"""
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


class StyleTransferNet(torch.nn.Module):
    """Neural network that implements the style transfer algorithm of Gatys"""
    vgg19_normalization_mean = torch.tensor([0.485, 0.456, 0.406]).to(get_device())
    vgg19_normalization_std = torch.tensor([0.229, 0.224, 0.225]).to(get_device())
    #content_layers_default = ['conv_4']
    content_layers_default = ['conv_10']  # Block 4, convolution 2
    #content_layers_default = ['relu_10']  # Block 4, convolution 2 (Gatys original)
    #style_layers_default = ['conv_1', 'conv_2', 'conv_3', 'conv_4', 'conv_5']
    style_layers_default = ['conv_1', 'conv_3', 'conv_5', 'conv_9', 'conv_13']  # First convolution from each block
    #style_layers_default = ['relu_1', 'relu_3', 'relu_5', 'relu_9', 'relu_13']  # First convolution from each block (Gatys original)

    def __init__(self, content_img, style_img, content_weight, style_weight, tv_weight, content_layers=None, style_layers=None):
        super(StyleTransferNet, self).__init__()

        # Check inputs
        if content_layers is None:
            content_layers = list(self.content_layers_default)
        else:
            assert isinstance(content_layers, Iterable) and all([isinstance(x, str) for x in content_layers]), ValueError(f"content_layers must be a list of layer names, got {content_layers}")
        if style_layers is None:
            style_layers = list(self.style_layers_default)
        else:
            assert isinstance(style_layers, Iterable) and all([isinstance(x, str) for x in style_layers]), ValueError(f"style_layers must be a list of layer names, got {style_layers}")

        # Initialize VGG19 reference architecture
        cnn = models.vgg19(pretrained=True).features.to(get_device()).eval()

        # Start with image normalization layer
        normalization = Normalization(self.vgg19_normalization_mean, self.vgg19_normalization_std)
        self.model = nn.Sequential(normalization)

        # Add TV loss layer
        self.tv_loss = TVLoss(tv_weight)
        self.model.add_module("tv_loss", self.tv_loss)

        # Add VGG19 layers, with added losses
        self.content_losses = []
        self.style_losses = []
        i = 0  # increment every time we see a conv
        for layer in cnn.children():
            if isinstance(layer, nn.Conv2d):
                i += 1
                name = 'conv_{}'.format(i)
            elif isinstance(layer, nn.ReLU):
                name = 'relu_{}'.format(i)
                # The in-place version doesn't play very nicely with the ContentLoss
                # and StyleLoss we insert below. So we replace with out-of-place
                # ones here.
                layer = nn.ReLU(inplace=False)
            elif isinstance(layer, nn.MaxPool2d):
                name = 'pool_{}'.format(i)
            elif isinstance(layer, nn.BatchNorm2d):
                name = 'bn_{}'.format(i)
            else:
                raise RuntimeError('Unrecognized layer: {}'.format(layer.__class__.__name__))

            self.model.add_module(name, layer)

            if name in content_layers:
                # add content loss:
                target = self.model(content_img).detach()
                content_loss = ContentLoss(target, content_weight)
                self.model.add_module("content_loss_{}".format(i), content_loss)
                self.content_losses.append(content_loss)

            if name in style_layers:
                # add style loss:
                target_feature = self.model(style_img).detach()
                style_loss = StyleLoss(target_feature, style_weight)
                self.model.add_module("style_loss_{}".format(i), style_loss)
                self.style_losses.append(style_loss)

        # now we trim off the layers after the last content and style losses
        for i in range(len(self.model) - 1, -1, -1):
            if isinstance(self.model[i], ContentLoss) or isinstance(self.model[i], StyleLoss):
                break

        self.model = self.model[:(i + 1)]

    def forward(self, x):
        return self.model(x)


class ScaleGradients(torch.autograd.Function):
    """Scale gradients in the backward pass"""
    @staticmethod
    def forward(self, input_tensor, strength):
        self.strength = strength
        return input_tensor

    @staticmethod
    def backward(self, grad_output):
        grad_input = grad_output.clone()
        grad_input = grad_input / (torch.norm(grad_input, keepdim=True) + 1e-8)
        return grad_input * self.strength, None


class ContentLoss(nn.Module):
    """Style transfer content loss: penalizes differences between a target image and generated image activations"""
    def __init__(self, target, content_weight):
        super(ContentLoss, self).__init__()
        self.target = target.detach()  # Do not compute gradients for target image
        self.content_weight = content_weight

    def forward(self, input):
        #self.loss = ScaleGradients.apply(F.mse_loss(input, self.target), self.content_weight) * self.content_weight
        # Note original Gatys paper loss is sum * 1/2 * weight, while this is mean * weight, but this implementation is compatible with jcjohnson
        self.loss = F.mse_loss(input, self.target) * self.content_weight
        return input


def gram_matrix(input):
    """Computes the Gram matrix of a single image after a convolution layer"""
    a, b, c, d = input.size()  # a=batch size(=1)
                               # b=number of feature maps (convolution channels)
                               # (c,d)=dimensions of a f. map (N=c*d)
    features = input.view(a * b, c * d)  # resise F_XL into \hat F_XL

    G = torch.mm(features, features.t())  # compute the gram product

    # we 'normalize' the values of the gram matrix
    # by dividing by the number of element in each feature maps.
    # Note in Gatys paper this normalization is done in the loss, not here, but it's equivalent
    return G.div(a * b * c * d)


class StyleLoss(nn.Module):
    """Style transfer style loss: penalizes differences between target image and generated image Gram matrices (activation correlations)"""
    def __init__(self, target, style_weight):
        super(StyleLoss, self).__init__()
        self.target = gram_matrix(target).detach()
        self.style_weight = style_weight

    def forward(self, input):
        G = gram_matrix(input)
        #self.loss = ScaleGradients.apply(F.mse_loss(G, self.target), self.style_weight) * self.style_weight
        # Note original Gatys paper loss is sum * 1/4 * weight, while this is mean * weight, but this implementation is compatible with jcjohnson
        self.loss = F.mse_loss(G, self.target) * self.style_weight
        return input


class TVLoss(nn.Module):
    """Total Variation loss: penalizes neighbouring with different values"""

    def __init__(self, strength):
        super(TVLoss, self).__init__()
        self.strength = strength

    def forward(self, input):
        self.x_diff = input[:,:,1:,:] - input[:,:,:-1,:]
        self.y_diff = input[:,:,:,1:] - input[:,:,:,:-1]
        self.loss = self.strength / torch.numel(input) * (torch.sum(torch.abs(self.x_diff)) + torch.sum(torch.abs(self.y_diff)))
        return input


class Normalization(nn.Module):
    """Layer that normalizes an image using given mean and standard deviation statistics"""
    def __init__(self, mean, std):
        super(Normalization, self).__init__()
        # .view the mean and std to make them [C x 1 x 1] so that they can
        # directly work with image Tensor of shape [B x C x H x W].
        # B is batch size. C is number of channels. H is height and W is width.
        self.mean = mean.clone().detach().view(-1, 1, 1)
        self.std = std.clone().detach().view(-1, 1, 1)

    def forward(self, img):
        # normalize img
        return (img - self.mean) / self.std



"""Main entrypoing for running style transfer as a python library

References:
    - Pytorch Neural Style tutorial: https://pytorch.org/tutorials/advanced/neural_style_tutorial.html
"""

from copy import deepcopy
import logging
import numpy as np
import torch
import torch.optim as optim
import torchvision.transforms as transforms

from .styletransfernet import StyleTransferNet, get_device


def _style_transfer_pass(content_img, style_img, num_steps=1000, style_weight=1e6, content_weight=1,
    tv_weight=1, output_resolution=None, init_img=None, content_layers=None, style_layers=None):
    """Runs a style transfer pass using Gatys method
    
    Arguments:
        content_img: original content image on which to apply style transfer. Image must be in PIL format.
        style_img: image with the style to transfer. Image must be in PIL format.
        num_steps: number of optimizer iterations to run
        style_weight: weight of the style loss in the optimization
        content_weight: weight of the content loss in the optimization
        tv_weight: weight of the Total Variation loss in the optimization
        output_resolution: desired resolution of the resultant image.
            Must be in format 'COLUMNS' or 'COLUMNSxROWS' (e.g. 640 or 640x480)
        init_img: image to initialize the optimization procedure. Must be a PIL image. If None, use content image.
        content_layers: iterable of VGG19 layers names into which to impose content losses. If None, a default choice is used.
        style_layers: iterable of VGG19 layers names into which to impose content losses. If None, a default choice is used.
    
    Returns a PIL image with the result of the style transfer.
    """
    # Format input images
    logging.info(f"Starting style transfer pass...")
    logging.info(f"Received content image size: {np.array(content_img).shape[:2]}")
    logging.info(f"Received style image size: {np.array(style_img).shape[:2]}")
    if init_img is not None:
        logging.info(f"Received initialization image size: {np.array(init_img).shape[:2]}")
    if output_resolution:
        logging.info(f"Preparing for output resolution {output_resolution}")
    content_img, style_img, init_img = _format_input_images(
        content_img,
        style_img,
        init_img,
        output_resolution
    )
    logging.info(f"Preprocessed content image size: {tuple(content_img.shape[2:])}")
    logging.info(f"Preprocessed style image size: {tuple(content_img.shape[2:])}")
    if init_img is not None:
        logging.info(f"Preprocessed initialization image size: {tuple(init_img.shape[2:])}")

    # Initialize style network
    style_network = StyleTransferNet(content_img, style_img, content_weight, style_weight, tv_weight, content_layers, style_layers).to(get_device())
    logging.info(f"StyleTransferNet created by the following architecture:\n{style_network.model}")

    # Initialize synthetic image
    input_img = deepcopy(init_img if init_img is not None else content_img).to(get_device())
    #input_img = deepcopy(init_img if init_img is not None else torch.randn(content_img.data.size())).to(get_device())

    # Prepare optimizer
    optimizer = optim.LBFGS([input_img.requires_grad_()], tolerance_grad=0, line_search_fn="strong_wolfe")
    #optimizer = optim.Adam([input_img.requires_grad_()])

    logging.info('Optimizing...')
    run = [0]
    while run[0] <= num_steps:
        def closure():
            # correct the values of updated input image
            # input_img.data.clamp_(0, 1)  # FIXME: not sure if this is helping. Produces training divergence

            optimizer.zero_grad()
            style_network(input_img)
            style_score = 0
            content_score = 0

            for sl in style_network.style_losses:
                style_score += sl.loss  # TODO: in Gatys the different styles losses are averaged, not summed
            for cl in style_network.content_losses:
                content_score += cl.loss
            tv_score = style_network.tv_loss.loss

            loss = style_score + content_score + tv_score
            loss.backward()

            run[0] += 1
            if run[0] % 50 == 0:
                logging.info(f'Step: {run[0]}, Style Loss: {style_score.item():6f}, Content Loss: {content_score.item():6f}, TV Loss: {tv_score.item():6f}, Total Loss: {loss.item():6f}')

            return style_score + content_score

        optimizer.step(closure)

    # a last correction...
    input_img.data.clamp_(0, 1)

    return _format_output_image(input_img)


def style_transfer(content_img, style_img, num_steps=1000, style_weight=1000000, content_weight=1,
    tv_weight=0, output_resolution=None, content_layers=None, style_layers=None, upscaling_rounds=1, 
    upscales_per_round=7, starting_resolution=256):
    """Transfers the style from one image to another.
    
    Arguments:
        content_img: original content image on which to apply style transfer. Image must be in PIL format.
        style_img: image with the style to transfer. Image must be in PIL format.
        num_steps: number of optimizer iterations to run
        style_weight: weight of the style loss in the optimization
        content_weight: weight of the content loss in the optimization
        tv_weight: weight of the Total Variation loss in the optimization
        output_resolution: desired resolution of the resultant image.
            Must be in format 'COLUMNS' or 'COLUMNSxROWS' (e.g. 640 or 640x480)
        init_img: image to initialize the optimization procedure. Must be a PIL image. If None, use content image.
        content_layers: iterable of VGG19 layers names into which to impose content losses. If None, a default choice is used.
        style_layers: iterable of VGG19 layers names into which to impose content losses. If None, a default choice is used.
        upscaling_rounds: number of upscaling rounds to perform. If 0 or None, the transfer is directly performed at
            target resolution, which is faster but yields poorer quality.
        upscales_per_round: number of upscaling steps to perform in each upscaling round.
        starting_resolution: resolution at which to start upscaling steps.
    
    Returns a PIL image with the result of the style transfer.
    """
    # If 0 or None rounds, just performe a simple pass
    if upscaling_rounds is None or upscaling_rounds == 0:
        return _style_transfer_pass(
            content_img, 
            style_img, 
            num_steps, 
            style_weight, 
            content_weight, 
            tv_weight,
            output_resolution=output_resolution,
            content_layers=content_layers,
            style_layers=style_layers
        )

    # Check arguments
    output_shape = _get_target_shape(content_img, output_resolution)
    if starting_resolution > output_shape[0]:
        logging.warning(f"Output resolution {output_shape[0]} is smaller than starting resolution {starting_resolution}, adjusting staring to half of output")
        starting_resolution = int(np.floor(output_shape[0] / 2))

    # Multiresolution rounds
    seed = None
    for round in range(upscaling_rounds):
        logging.info(f"Upscaling strategy round {round}")
        iterations = num_steps
        resolutions = np.linspace(starting_resolution, output_shape[0], upscales_per_round, dtype=int)
        for stepnumber, res in enumerate(resolutions):
            logging.info(f"Upscaling round {round} step {stepnumber}: upscaling to resolution {res}")
            seed = _style_transfer_pass(
                content_img, 
                style_img, 
                iterations, 
                style_weight, 
                content_weight, 
                tv_weight,
                output_resolution=res,
                init_img=seed,
                content_layers=content_layers,
                style_layers=style_layers
            )
            # iterations = max(iterations / 2, 100)

    return seed


def _get_target_shape(content_img, output_resolution=None):
    """Returns the shape that the resultant image will take for given content and desired output resolution"""
    if output_resolution is not None:
        try:
            if isinstance(output_resolution, np.integer):
                output_shape = [output_resolution]
            else:
                output_shape = [int(x) for x in output_resolution.split("x")]
            assert 1 <= len(output_shape) <= 2
            if len(output_shape) == 1:
                output_columns = output_shape[0]
                content_columns, content_rows, _ = np.array(content_img).shape
                output_rows = int(np.floor(output_columns / content_columns * content_rows))
                output_shape = [output_columns, output_rows]
        except:
            raise ValueError("Output resolution must be in format 'COLUMNS' or 'COLUMNSxROWS' (e.g. 640 or 640x480")
    else:
        output_shape = np.array(content_img).shape[0:2]
    return output_shape


def _format_input_images(content, style, init=None, output_resolution=None):
    """Scale and format input PIL images for correct performance of the Style Transfer method.
    
    This means:
        - Transform PIL images to pytorch tensors
        - Content image is rescaled to desired output resolution
        - Style image is rescaled to the same dimensions as the content image
        - Initialization image (if provided) is rescaled to the same dimensions as the content image
    """
    # Content tranformations
    output_shape = _get_target_shape(content, output_resolution)
    if output_shape != content.size:
        content_transforms = transforms.Compose([transforms.ToTensor(), transforms.Resize(output_shape)])
    else:
        content_transforms = transforms.ToTensor()
    # Style transformations
    style_transforms = transforms.Compose([transforms.ToTensor(), transforms.Resize(output_shape)])
    # Initialization image transformations (if provided)
    if init is not None:
        init_transforms = transforms.Compose([transforms.ToTensor(), transforms.Resize(output_shape)])
    # Apply and return
    return (
        content_transforms(content).unsqueeze(0).to(get_device()),
        style_transforms(style).unsqueeze(0).to(get_device()),
        init_transforms(init).unsqueeze(0).to(get_device()) if init is not None else None
    )

def _format_output_image(output):
    """Recovers the output of the style transfer as a PIL image"""
    output_transforms = transforms.ToPILImage()
    return output_transforms(output.detach().squeeze(0))



import os
from setuptools import setup


# Read long description from readme
with open("README.md", "r", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()


# Get tag from Github environment variables
TAG = os.environ['GITHUB_TAG'] if 'GITHUB_TAG' in os.environ else None


setup(
    name="guertena",
    version=TAG if TAG is not None else "develop",
    description="Guertena is an easy to use, quality oriented python library for neural style transfer.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=['guertena'],
    install_requires=[
        'numpy>=1.20,<2',
        'torch>=1.9,<2',
        'torchvision>=0.10,<1'
    ],
    author="Alvaro Barbero",
    url='https://github.com/albarji/guertena',
    license='MIT',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: GPU :: NVIDIA CUDA',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Artistic Software',
        'Topic :: Multimedia :: Graphics'
    ],
    keywords='artificial-intelligence, deep-learning, neural-style-transfer',
    test_suite="pytest",
)